# pyflyby/_livepatch.py
# Copyright (C) 2011, 2012, 2013, 2014, 2015 Karl Chen.

r"""
livepatch/xreload: Alternative to reload().

xreload performs a "live patch" of the modules/classes/functions/etc that have
already been loaded in memory.  It does so by executing the module in a
scratch namespace, and then patching classes, methods and functions in-place.
New objects are copied into the target namespace.

This addresses cases where one module imported functions from another
module.

For example, suppose m1.py contains::

  from m2 import foo
  def print_foo():
      return foo()

and m2.py contains::

  def foo():
      return 42

If you edit m2.py and modify ``foo``, then reload(m2) on its own would not do
what you want.  You would also need to reload(m1) after reload(m2).  This is
because the built-in reload affects the module being reloaded, but references
to the old module remain.  On the other hand, xreload() patches the existing
m2.foo, so that live references to it are updated.

In table form::

  Undesired effect:  reload(m2)
  Undesired effect:  reload(m1);  reload(m2)
  Desired effect:    reload(m2);  reload(m1)

  Desired effect:   xreload(m2)
  Desired effect:   xreload(m1); xreload(m2)
  Desired effect:   xreload(m2); xreload(m1)

Even with just two modules, we can see that xreload() is an improvement.  When
working with a large set of interdependent modules, it becomes infeasible to
know the precise sequence of reload() calls that would be necessary.
xreload() really shines in that case.

This implementation of xreload() was originally based the following
mailing-list post by Guido van Rossum:

    https://mail.python.org/pipermail/edu-sig/2007-February/007787.html

Customizing behavior
====================

If a class/function/module/etc has an attribute __livepatch__, then this
function is called *instead* of performing the regular livepatch mechanism.

The __livepatch__() function is called with the following arguments:

  - ``old``         : The object to be updated with contents of ``new``
  - ``new``         : The object whose contents to put into ``old``
  - ``do_livepatch``: A function that can be called to do the standard
                      livepatch, replacing the contents of ``old`` with ``new``.
                      If it's not possible to livepatch ``old``, it returns
                      ``new``.  The ``do_livepatch`` function takes no arguments.
                      Calling the ``do_livepatch`` function is roughly
                      equivalent to calling ``pyflyby.livepatch(old, new,
                      modname=modname, heed_hook=False)``.
  - ``modname``     : The module currently being updated.  Recursively called
                      updates should keep track of the module being updated to
                      avoid touching other modules.

These arguments are matched by *name* and are passed only if the
``__livepatch__`` function is declared to take such named arguments or it takes
\**kwargs.  If the ``__livepatch__`` function takes \**kwargs, it should ignore
unknown arguments, in case new parameters are added in the future.

If the object being updated is an object instance, and ``__livepatch__`` is a
method, then the function is bound to the new object, i.e. the ``self``
parameter is the same as ``new``.

If the ``__livepatch__`` function successfully patched the ``old`` object, then
it should return ``old``.  If it is unable to patch, it should return ``new``.

Examples
--------

By default, any attributes on an existing function are updated with ones from
the new function.  If you want a memoized function to keep its cache across
xreload, you could implement that like this::

  def memoize(function):
      cache = {}
      def wrapped_fn(*args):
          try:
              return cache[args]
          except KeyError:
              result = function(*args)
              cache[args] = result
              return result
      wrapped_fn.cache = cache
      def my_livepatch(old, new, do_livepatch):
          keep_cache = dict(old.cache)
          result = do_livepatch()
          result.cache.update(keep_cache)
          return result
      wrapped_fn.__livepatch__ = my_livepatch
      return wrapped_fn

XXX change example b/c cache is already cleared by default
XXX maybe global cache

  class MyObj(...):
      def __livepatch__(self, old):
          self.__dict__.update(old.__dict__)
          return self

  class MyObj(...):
      def __init__(self):
          self._my_cache = {}

      def __livepatch__(self, old, do_livepatch):
          keep_cache = dict(old._my_cache)
          result = do_livepatch()
          result._my_cache.update(keep_cache)
          return result

XXX test

"""



import ast
import os
import re
import sys
import time
import types

from   importlib                import reload as reload_module

import inspect
from   pyflyby._log             import logger


# Keep track of when the process was started.
if os.uname()[0] == 'Linux':
    _PROCESS_START_TIME = os.stat("/proc/%d"%os.getpid()).st_ctime
else:
    try:
        import psutil
    except ImportError:
        # Todo: better fallback
        _PROCESS_START_TIME = time.time()
    else:
        _PROCESS_START_TIME = psutil.Process(os.getpid()).create_time()


class UnknownModuleError(ImportError):
    pass


def livepatch(old, new, modname=None,
              visit_stack=(), cache=None, assume_type=None,
              heed_hook=True):
    """
    Livepatch ``old`` with contents of ``new``.

    If ``old`` can't be livepatched, then return ``new``.

    :param old:
      The object to be updated
    :param new:
      The object used as the source for the update.
    :type modname:
      ``str``
    :param modname:
      Only livepatch ``old`` if it was defined in the given fully-qualified
      module name.  If ``None``, then update regardless of module.
    :param assume_type:
      Update as if both ``old`` and ``new`` were of type ``assume_type``.  If
      ``None``, then ``old`` and ``new`` must have the same type.
      For internal use.
    :param cache:
      Cache of already-updated objects.  Map from (id(old), id(new)) to result.
    :param visit_stack:
      Ids of objects that are currently being updated.
      Used to deal with reference cycles.
      For internal use.
    :param heed_hook:
      If ``True``, heed the ``__livepatch__`` hook on ``new``, if any.
      If ``False``, ignore any ``__livepatch__`` hook on ``new``.
    :return:
      Either live-patched ``old``, or ``new``.
    """
    if old is new:
        return new
    # If we're already visiting this object (due to a reference cycle), then
    # don't recurse again.
    if id(old) in visit_stack:
        return old
    if cache is None:
        cache = {}
    cachekey = (id(old), id(new))
    try:
        return cache[cachekey]
    except KeyError:
        pass
    visit_stack += (id(old),)
    def do_livepatch():
        new_modname = _get_definition_module(new)
        if modname and new_modname and new_modname != modname:
            # Ignore objects that have been imported from another module.
            # Just update their references.
            return new
        if assume_type is not None:
            use_type = assume_type
        else:
            oldtype = type(old)
            newtype = type(new)
            if oldtype is newtype:
                # Easy, common case: Type didn't change.
                use_type = oldtype
            elif (oldtype.__name__ == newtype.__name__ and
                  oldtype.__module__ == newtype.__module__ == modname and
                  getattr(sys.modules[modname],
                          newtype.__name__, None) is newtype and
                  oldtype is livepatch(
                      oldtype, newtype, modname=modname,
                      visit_stack=visit_stack, cache=cache)):
                # Type of this object was defined in this module.  This
                # includes metaclasses defined in the same module.
                use_type = oldtype
            else:
                # If the type changed, then give up.
                return new
        try:
            mro = type.mro(use_type)
        except TypeError:
            mro = [use_type, object] # old-style class
        # Dispatch on type.  Include parent classes (in C3 linearized
        # method resolution order), in particular so that this works on
        # classes with custom metaclasses that subclass ``type``.
        for t in mro:
            try:
                update = _LIVEPATCH_DISPATCH_TABLE[t]
                break
            except KeyError:
                pass
        else:
            # We should have found at least ``object``
            raise AssertionError("unreachable")
        # Dispatch.
        return update(old, new, modname=modname,
                      cache=cache, visit_stack=visit_stack)
    if heed_hook:
        hook = (getattr(new, "__livepatch__", None) or
                getattr(new, "__reload_update__", None))
        # XXX if unbound method or a descriptor, then we should ignore it.
        # XXX test for that.
    else:
        hook = None
    if hook is None:
        # No hook is defined or the caller instructed us to ignore it.
        # Do the standard livepatch.
        result = do_livepatch()
    else:
        # Call a hook for updating.
        # Build dict of optional kwargs.
        avail_kwargs = dict(
            old=old,
            new=new,
            do_livepatch=do_livepatch,
            modname=modname,
            cache=cache,
            visit_stack=visit_stack)
        # Find out which optional kwargs the hook wants.
        kwargs = {}
        argspec = inspect.getfullargspec(hook)
        argnames = argspec.args
        if hasattr(hook, "__func__"):
            # Skip 'self' arg.
            argnames = argnames[1:]
        # Pick kwargs that are wanted and available.
        args = []
        kwargs = {}
        for n in argnames:
            try:
                kwargs[n] = avail_kwargs[n]
                if argspec.varkw:
                    break
            except KeyError:
                # For compatibility, allow first argument to be 'old' with any
                # name, as long as there's no other arg 'old'.
                # We intentionally allow this even if the user specified
                # **kwargs.
                if not args and not kwargs and 'old' not in argnames:
                    args.append(old)
                else:
                    # Rely on default being set.  If a default isn't set, the
                    # user will get a TypeError.
                    pass
        if argspec.varkw:
            # Use all available kwargs.
            kwargs = avail_kwargs
        # Call hook.
        result = hook(*args, **kwargs)
    cache[cachekey] = result
    return result


def _livepatch__module(old_mod, new_mod, modname, cache, visit_stack):
    """
    Livepatch a module.
    """
    result = livepatch(old_mod.__dict__, new_mod.__dict__,
                       modname=modname,
                       cache=cache, visit_stack=visit_stack)
    assert result is old_mod.__dict__
    return old_mod


def _livepatch__dict(old_dict, new_dict, modname, cache, visit_stack):
    """
    Livepatch a dict.
    """
    oldnames = set(old_dict)
    newnames = set(new_dict)
    # Add newly introduced names.
    for name in newnames - oldnames:
        old_dict[name] = new_dict[name]
    # Delete names that are no longer current.
    for name in oldnames - newnames:
        del old_dict[name]
    # Livepatch existing entries.
    updated_names = sorted(oldnames & newnames, key=str)
    for name in updated_names:
        old = old_dict[name]
        updated = livepatch(old, new_dict[name],
                            modname=modname,
                            cache=cache, visit_stack=visit_stack)
        if updated is not old:
            old_dict[name] = updated
    return old_dict


def _livepatch__function(old_func, new_func, modname, cache, visit_stack):
    """
    Livepatch a function.
    """
    # If the name differs, then don't update the existing function - this
    # is probably a reassigned function.
    if old_func.__name__ != new_func.__name__:
        return new_func
    # Check if the function's closure is compatible.  If not, then return the
    # new function without livepatching.  Note that cell closures can't be
    # modified; we can only livepatch cell values.
    old_closure = old_func.__closure__ or ()
    new_closure = new_func.__closure__ or ()
    if len(old_closure) != len(new_closure):
        return new_func
    if old_func.__code__.co_freevars != new_func.__code__.co_freevars:
        return new_func
    for oldcell, newcell in zip(old_closure, new_closure):
        oldcellv = oldcell.cell_contents
        newcellv = newcell.cell_contents
        if type(oldcellv) != type(newcellv):
            return new_func
        if isinstance(oldcellv, (
                types.FunctionType, types.MethodType, type, dict)):
            # Updateable type.  (Todo: make this configured globally.)
            continue
        try:
            if oldcellv is newcellv or oldcellv == newcellv:
                continue
        except Exception:
            pass
        # Non-updateable and not the same as before.
        return new_func
    # Update function code, defaults, doc.
    old_func.__code__ = new_func.__code__
    old_func.__defaults__ = new_func.__defaults__
    old_func.__doc__ = new_func.__doc__
    # Update dict.
    livepatch(old_func.__dict__, new_func.__dict__,
              modname=modname, cache=cache, visit_stack=visit_stack)
    # Update the __closure__.  We can't set __closure__ because it's a
    # read-only attribute; we can only livepatch its cells' values.
    for oldcell, newcell in zip(old_closure, new_closure):
        oldcellv = oldcell.cell_contents
        newcellv = newcell.cell_contents
        livepatch(oldcellv, newcellv,
                  modname=modname, cache=cache, visit_stack=visit_stack)
    return old_func


def _livepatch__method(old_method, new_method, modname, cache, visit_stack):
    """
    Livepatch a method.
    """
    _livepatch__function(old_method.__func__, new_method.__func__,
                         modname=modname,
                         cache=cache, visit_stack=visit_stack)
    return old_method


def _livepatch__setattr(oldobj, newobj, name, modname, cache, visit_stack):
    """
    Livepatch something via setattr, i.e.::

       oldobj.{name} = livepatch(oldobj.{name}, newobj.{name}, ...)
    """
    newval = getattr(newobj, name)
    assert type(newval) is not types.MemberDescriptorType
    try:
        oldval = getattr(oldobj, name)
    except AttributeError:
        # This shouldn't happen, but just ignore it.
        setattr(oldobj, name, newval)
        return
    # If it's the same object, then skip.  Note that if even if 'newval ==
    # oldval', as long as they're not the same object instance, we still
    # livepatch.  We want mutable data structures get livepatched instead of
    # replaced.  Avoiding calling '==' also avoids the risk of user code
    # having defined '==' to do something unexpected.
    if newval is oldval:
        return
    # Livepatch the member object.
    newval = livepatch(
        oldval, newval, modname=modname, cache=cache, visit_stack=visit_stack)
    # If the livepatch succeeded then we don't need to setattr.  It should be
    # a no-op but we avoid it just to minimize any chance of setattr causing
    # problems in corner cases.
    if newval is oldval:
        return
    # Livepatch failed, so we have to update the container with the new member
    # value.
    setattr(oldobj, name, newval)


def _livepatch__class(oldclass, newclass, modname, cache, visit_stack):
    """
    Livepatch a class.

    This is similar to _livepatch__dict(oldclass.__dict__, newclass.__dict__).
    However, we can't just operate on the dict, because class dictionaries are
    special objects that don't allow setitem, even though we can setattr on
    the class.
    """
    # Collect the names to update.
    olddict = oldclass.__dict__
    newdict = newclass.__dict__
    # Make sure slottiness hasn't changed -- i.e. if class was changed to have
    # slots, or changed to not have slots, or if the slot names changed in any
    # way, then we can't livepatch the class.
    # Note that this is about whether instances of this class are affected by
    # __slots__ or not.  The class type itself will always use a __dict__.
    if olddict.get("__slots__") != newdict.get("__slots__"):
        return newclass
    oldnames = set(olddict)
    newnames = set(newdict)
    for name in oldnames - newnames:
        delattr(oldclass, name)
    for name in newnames - oldnames:
        setattr(oldclass, name, newdict[name])
    oldclass.__bases__ = newclass.__bases__
    names = oldnames & newnames
    names.difference_update(olddict.get("__slots__", []))
    names.discard("__slots__")
    names.discard("__dict__")
    # Python < 3.3 doesn't support modifying __doc__ on classes with
    # non-custom metaclasses.  Attempt to do it and ignore failures.
    # http://bugs.python.org/issue12773
    names.discard("__doc__")
    try:
        oldclass.__doc__ = newclass.__doc__
    except AttributeError:
        pass
    # Loop over attributes to be updated.
    for name in sorted(names):
        _livepatch__setattr(
            oldclass, newclass, name, modname, cache, visit_stack)
    return oldclass


def _livepatch__object(oldobj, newobj, modname, cache, visit_stack):
    """
    Livepatch a general object.
    """
    # It's not obvious whether ``oldobj`` and ``newobj`` are actually supposed
    # to represent the same object.  For now, we take a middle ground of
    # livepatching iff the class was also defined in the same module.  In that
    # case at least we know that the object was defined in this module and
    # therefore more likely that we should livepatch.
    if modname and _get_definition_module(type(oldobj)) != modname:
        return newobj
    if hasattr(type(oldobj), "__slots__"):
        assert oldobj.__slots__ == newobj.__slots__
        for name in newobj.__slots__:
            hasold = hasattr(oldobj, name)
            hasnew = hasattr(newobj, name)
            if hasold and hasnew:
                _livepatch__setattr(oldobj, newobj, name,
                                          modname, cache, visit_stack)
            elif hasold and not hasnew:
                delattr(oldobj, name)
            elif not hasold and hasnew:
                setattr(oldobj, getattr(newobj, name))
            elif not hasold and not hasnew:
                pass
            else:
                raise AssertionError
        return oldobj
    elif type(getattr(oldobj, "__dict__", None)) is dict:
        livepatch(
            oldobj.__dict__, newobj.__dict__,
            modname=modname, cache=cache, visit_stack=visit_stack)
        return oldobj
    else:
        return newobj

_LIVEPATCH_DISPATCH_TABLE = {
    object            : _livepatch__object,
    dict              : _livepatch__dict,
    type              : _livepatch__class,
    types.FunctionType: _livepatch__function,
    types.MethodType  : _livepatch__method,
    types.ModuleType  : _livepatch__module,
}


def _get_definition_module(obj):
    """
    Get the name of the module that an object is defined in, or ``None`` if
    unknown.

    For classes and functions, this returns the ``__module__`` attribute.

    For object instances, this returns ``None``, ignoring the ``__module__``
    attribute.  The reason is that the ``__module__`` attribute on an instance
    just gives the module that the class was defined in, which is not
    necessarily the module where the instance was constructed.

    :rtype:
      ``str``
    """
    if isinstance(obj, (type, types.FunctionType,
                        types.MethodType)):
        return getattr(obj, "__module__", None)
    else:
        return None


def _format_age(t):
    secs = time.time() - t
    if secs > 120:
        return "%dm%ds" %(secs//60, secs%60)
    else:
        return "%ds" %(secs,)


def _interpret_module(arg):
    def mod_fn(module):
        return getattr(module, "__file__", None)

    if isinstance(arg, str):
        try:
            return sys.modules[arg]
        except KeyError:
            pass
        if arg.startswith("/"):
            fn = os.path.realpath(arg)
            if fn.endswith(".pyc") or fn.endswith(".pyo"):
                fn = fn[:-1]
            if fn.endswith(".py"):
                relevant_fns = set([fn, fn+"c", fn+"o"])
            else:
                relevant_fns = set([fn])
            found_modules = [
                m for _,m in sorted(sys.modules.items())
                if os.path.realpath(mod_fn(m) or "/") in relevant_fns ]
            if not found_modules:
                raise UnknownModuleError(
                    "No loaded module uses path %s" % (fn,))
            if len(found_modules) > 1:
                raise UnknownModuleError(
                    "Multiple loaded modules use path %s: %r"
                    % (fn, found_modules))
            return found_modules[0]
        if arg.endswith(".py") and "/" not in arg:
            name = arg[:-3]
            relevant_bns = set([arg, arg+"c", arg+"o"])
            found_modules = [
                m for n,m in sorted(sys.modules.items())
                if (n==name or
                    os.path.basename(mod_fn(m) or "/") in relevant_bns)]
            if not found_modules:
                raise UnknownModuleError(
                    "No loaded module named %s" % (name,))
            if len(found_modules) > 1:
                raise UnknownModuleError(
                    "Multiple loaded modules named %s: %r"
                    % (name, found_modules))
            return found_modules[0]
        raise UnknownModuleError(arg)
    if isinstance(arg, types.ModuleType):
        return arg
    try:
        # Allow fake modules.
        if sys.modules[arg.__name__] is arg:
            return arg
    except Exception:
        pass
    raise TypeError("Expected module, module name, or filename; got %s"
                    % (type(arg).__name__))


def _xreload_module(module, filename, force=False):
    """
    Reload a module in place, using livepatch.

    :type module:
      ``ModuleType``
    :param module:
      Module to reload.
    :param force:
      Whether to reload even if the module has not been modified since the
      previous load.  If ``False``, then do nothing.  If ``True``, then reload.
    """
    import linecache
    if not filename or not filename.endswith(".py"):
        # If there's no *.py source file for this module, then fallback to
        # built-in reload().
        return reload_module(module)
    # Compare mtime of the file with the load time of the module.  If the file
    # wasn't touched, we don't need to do anything.
    try:
        mtime = os.stat(filename).st_mtime
    except OSError:
        logger.info("Can't find %s", filename)
        return None
    if not force:
        try:
            old_loadtime = module.__loadtime__
        except AttributeError:
            # We only have a __loadtime__ attribute if we were the ones that
            # loaded it.  Otherwise, fall back to the process start time as a
            # conservative bound.
            old_loadtime = _PROCESS_START_TIME
        if old_loadtime > mtime:
            logger.debug(
                "NOT reloading %s (file %s modified %s ago but loaded %s ago)",
                module.__name__, filename, _format_age(mtime),
                _format_age(old_loadtime))
            return None
        # Keep track of previously imported source.  If the file's timestamp
        # was touched, but the content unchanged, we can avoid reloading.
        cached_lines = linecache.cache.get(filename, (None,None,None,None))[2]
    else:
        cached_lines = None
    # Re-read source for module from disk, and update the linecache.
    source = ''.join(linecache.updatecache(filename))
    # Skip reload if the content didn't change.
    if cached_lines is not None and source == ''.join(cached_lines):
        logger.debug(
            "NOT reloading %s (file %s touched %s ago but content unchanged)",
            module.__name__, filename, _format_age(mtime))
        return module
    logger.info("Reloading %s (modified %s ago) from %s",
                module.__name__, _format_age(mtime), filename)
    # Compile into AST.  We do this as a separate step from compiling to byte
    # code so that we can get the module docstring.
    astnode = compile(source, filename, "exec", ast.PyCF_ONLY_AST, 1)
    # Get the new docstring.
    try:
        if sys.versin_info > (3,10):
            doc = astnode.body[0].value.value
        else:
            doc = astnode.body[0].value.s
    except (AttributeError, IndexError):
        doc = None
    # Compile into code.
    code = compile(astnode, filename, "exec", 0, 1)
    # Execute the code.  We do so in a temporary namespace so that if this
    # fails, nothing changes.  It's important to set __name__ so that relative
    # imports work correctly.
    new_mod = types.ModuleType(module.__name__)
    new_mod.__file__ = filename
    new_mod.__doc__ = doc
    if hasattr(module, "__path__"):
        new_mod.__path__ = module.__path__
    MISSING = object()
    saved_mod = sys.modules.get(module.__name__, MISSING)
    try:
        # Temporarily put the temporary module in sys.modules, in case the
        # code references sys.modules[__name__] for some reason.  Normally on
        # success, we will revert this what that was there before (which
        # normally should be ``module``).  If an error occurs, we'll also
        # revert.  If the user has defined a __livepatch__ hook at the module
        # level, it's possible for result to not be the old module.
        sys.modules[module.__name__] = new_mod
        # *** Execute new code ***
        exec(code, new_mod.__dict__)
        # Normally ``module`` is of type ``ModuleType``.  However, in some
        # cases, the module might have done a "proxy module" trick where the
        # module is replaced by a proxy object of some other type.  Regardless
        # of the actual type, we do the update as ``module`` were of type
        # ``ModuleType``.
        assume_type = types.ModuleType
        # Livepatch the module.
        result = livepatch(module, new_mod, module.__name__,
                           assume_type=assume_type)
        sys.modules[module.__name__] = result
    except:
        # Either the module failed executing or the livepatch failed.
        # Revert to previous state.
        # Note that this isn't perfect because it's possible that the module
        # modified some global state in other modules.
        if saved_mod is MISSING:
            del sys.modules[module.__name__]
        else:
            sys.modules[module.__name__] = saved_mod
        raise
    # Update the time we last loaded the module.  We intentionally use mtime
    # here instead of time.time().  If we are on NFS, it's possible for the
    # filer's mtime and time.time() to not be synchronized.  We will be
    # comparing to mtime next time, so if we use only mtime, we'll be fine.
    module.__loadtime__ = mtime
    return module


def _get_module_py_file(module):
    filename = getattr(module, "__file__", None)
    if not filename:
        return None
    filename = re.sub("[.]py[co]$", ".py", filename)
    return filename


def xreload(*args):
    """
    Reload module(s).

    This function is more useful than the built-in reload().  xreload() uses a
    "live patch" approach that modifies existing functions, classes, and
    objects in-place.

    This addresses cases where one module imported functions from another
    module.

    For example, suppose m1.py contains::

      from m2 import foo
      def print_foo():
          return foo()

    and m2.py contains::

      def foo():
          return 42

    If you edit m2.py and modify ``foo``, then reload(m2) on its own would not
    do what you want.  The built-in reload affects the module being reloaded,
    but references to the old module remain.  On the other hand, xreload()
    patches the existing m2.foo, so that live references to it are updated.

    :type args:
      ``str`` s and/or ``ModuleType`` s
    :param args:
      Module(s) to reload.  If no argument is specified, then reload all
      recently modified modules.
    """
    if not args:
        for name, module in sorted(sys.modules.items()):
            if name == "__main__":
                continue
            filename = _get_module_py_file(module)
            if not filename:
                continue
            _xreload_module(module, filename)
        return
    # Treat xreload(list_of_module) like xreload(*list_of_modules).  We
    # intentionally do this after the above check so that xreload([]) does
    # nothing.
    if len(args) == 1 and isinstance(args[0], (tuple, list)):
        args = args[0]
    for arg in args:
        module = _interpret_module(arg)
        # Get the *.py filename for this module.
        filename = _get_module_py_file(module)
        # Reload the module.
        _xreload_module(module, filename)



# To remove imports from the set of known imports, write, for example:
#   __forget_imports__ = [
#      'numpy.sin',
#   ]
#
# You can write this in any *.py file in your $PYFLYBY_PATH,
# e.g. ~/.pyflyby/forget.py.
#
# This can be useful if you're inheriting somebody else's import database.
# You can inherit most of the parent database, but exclude certain imports.



"""
pyflyby/_saveframe.py

Provides a utility to save the info for debugging / reproducing any issue.
Checkout the doc of the `saveframe` function for more info.
"""

from __future__ import annotations

from   contextlib               import contextmanager
from   dataclasses              import dataclass
from   enum                     import Enum
import inspect
import keyword
import linecache
import logging
import os
import pickle
import re
import stat
import sys
import traceback

"""
The protocol used while pickling the frame's data.
"""
PICKLE_PROTOCOL=5

"""
The permissions used to create the file where the data is stored by the
'saveframe' utility.
"""
FILE_PERMISSION = 0o644

"""
The default filename used for storing the data when the user does not explicitly
provide one.
"""
DEFAULT_FILENAME = 'saveframe.pkl'


@dataclass
class ExceptionInfo:
    """
    A dataclass to store the exception info.
    """
    exception_string: str
    exception_full_string: str
    exception_class_name: str
    exception_class_qualname: str
    exception_object: object
    traceback: list


@dataclass
class FrameMetadata:
    """
    A dataclass to store a frame's metadata.
    """
    frame_index: int
    filename: str
    lineno: int
    function_name: str
    function_qualname: str
    function_object: bytes
    module_name: str
    code: str
    frame_identifier: str


class FrameFormat(Enum):
    """
    Enum class to store the different formats supported by the `frames` argument
    in the `saveframe` utility. See the doc of `saveframe` for more info.
    """
    NUM = "NUM"
    LIST = "LIST"
    RANGE = "RANGE"


def _get_saveframe_logger():
    """
    Get the logger used for the saveframe utility.
    """
    log_format = (
        "[%(asctime)s:%(msecs)03d pyflyby.saveframe:%(lineno)s "
        "%(levelname)s] %(message)s")
    log_datefmt = "%Y%m%d %H:%M:%S"
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.INFO)
    if not logger.handlers:
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(logging.INFO)
        formatter = logging.Formatter(
            fmt=log_format, datefmt=log_datefmt)
        stream_handler.setFormatter(formatter)
        logger.addHandler(stream_handler)
    return logger

_SAVEFRAME_LOGGER = _get_saveframe_logger()


@contextmanager
def _open_file(filename, mode):
    """
    A context manager to open the ``filename`` with ``mode``.
    This function ignores the ``umask`` while creating the file.

    :param filename:
      The file to open.
    :param mode:
      Mode in which to open the file.
    """
    old_umask = os.umask(0)
    fd = None
    file_obj = None
    try:
        fd = os.open(filename, os.O_WRONLY | os.O_CREAT | os.O_TRUNC, FILE_PERMISSION)
        file_obj = os.fdopen(fd, mode)
        yield file_obj
    finally:
        if file_obj is not None:
            file_obj.close()
        elif fd is not None:
            os.close(fd)
        os.umask(old_umask)


def _get_exception_info(exception_obj):
    """
    Get the metadata information for the ``exception_obj``.

    :param exception_obj:
      The exception raised by the user's code.
    :return:
      An `ExceptionInfo` object.
    """
    try:
        tb = (
            traceback.format_exception(
                type(exception_obj), exception_obj, exception_obj.__traceback__))
    except Exception as err:
        _SAVEFRAME_LOGGER.warning(
            "Error while formatting the traceback. Error: %a", err)
        tb = "Traceback couldn't be formatted"
    exception_info = ExceptionInfo(
        exception_string=str(exception_obj),
        exception_full_string=f'{exception_obj.__class__.__name__}: {exception_obj}',
        exception_class_name=exception_obj.__class__.__name__,
        exception_class_qualname=exception_obj.__class__.__qualname__,
        exception_object=exception_obj,
        traceback=tb
    )
    return exception_info


def _get_qualname(frame):
    """
    Get fully qualified name of the function for the ``frame``.

    In python 3.10, ``co_qualname`` attribute is not present, so use ``co_name``.
    """
    return (frame.f_code.co_qualname if hasattr(frame.f_code, "co_qualname")
            else frame.f_code.co_name)

def _get_frame_repr(frame):
    """
    Construct repr for the ``frame``. This is used in the info messages.

    :param frame:
      The frame object.
    :return:
      The string f'File: {filename}, Line: {lineno}, Function: {function_qualname}'
    """
    return (f"'File: {frame.f_code.co_filename}, Line: {frame.f_lineno}, "
            f"Function: {_get_qualname(frame)}'")


def _get_frame_local_variables_data(frame, variables, exclude_variables):
    """
    Get the local variables data of the ``frame``.

    :param frame:
      The frame object
    :param variables:
      Local variables to be included.
    :param exclude_variables:
      Local variables to be excluded.
    :return:
      A dict containing the local variables data, with the key as the variable
      name and the value as the pickled local variable value.
    """
    # A dict to store the local variables to be saved.
    local_variables_to_save = {}
    all_local_variables = frame.f_locals
    for variable in all_local_variables:
        # Discard the variables that starts with '__' like '__eq__', etc., to
        # keep the data clean.
        if variable.startswith('__'):
            continue
        if variables and variable not in variables:
            continue
        if exclude_variables and variable in exclude_variables:
            continue
        try:
            pickled_value = pickle.dumps(
                all_local_variables[variable], protocol=PICKLE_PROTOCOL)
        except Exception as err:
            _SAVEFRAME_LOGGER.warning(
                "Cannot pickle variable: %a for frame: %s. Error: %a. Skipping "
                "this variable and continuing.",
                variable, _get_frame_repr(frame), err)
        else:
            local_variables_to_save[variable] = pickled_value
    return local_variables_to_save


def _get_frame_function_object(frame):
    """
    Get the function object of the frame.

    This helper does a best-effort attempt to find the function object using
    locals and globals dict.

    :param frame:
      The frame object.
    :return:
      The function object from which the ``frame`` is originating.
    """
    func_name = frame.f_code.co_name
    func_qualname = _get_qualname(frame)
    info_msg = f"Can't get function object for frame: {_get_frame_repr(frame)}"
    return_msg = "Function object not found"
    # The function is most-likely either a local function or a class method.
    if func_name != func_qualname:
        prev_frame = frame.f_back
        # Handle the local functions.
        if "<locals>" in func_qualname:
            if prev_frame is None:
                _SAVEFRAME_LOGGER.info(info_msg)
                return return_msg
            # The function is present in the previous frame's (the parent) locals.
            if func_name in prev_frame.f_locals:
                return prev_frame.f_locals[func_name]
            _SAVEFRAME_LOGGER.info(info_msg)
            return return_msg
        # Handle the class methods.
        else:
            try:
                func_parent = func_qualname.split('.')[-2]
            except IndexError:
                _SAVEFRAME_LOGGER.info(info_msg)
                return return_msg
            # The parent is present in the globals, so extract the function object
            # using getattr.
            if func_parent in frame.f_globals:
                func_parent_obj = frame.f_globals[func_parent]
                if hasattr(func_parent_obj, func_name):
                    return getattr(func_parent_obj, func_name)
            _SAVEFRAME_LOGGER.info(info_msg)
            return return_msg
    # The function is most-likely a global function.
    else:
        if func_name in frame.f_globals:
            return frame.f_globals[func_name]
        _SAVEFRAME_LOGGER.info(info_msg)
        return return_msg


def _get_frame_module_name(frame):
    """
    Get the module name of the ``frame``.

    :param frame:
      The frame object.
    :return:
      The name of the module from which the ``frame`` is originating.
    """
    try:
        frame_module = inspect.getmodule(frame)
        if frame_module is not None:
            return frame_module.__name__
        _SAVEFRAME_LOGGER.info(
            "No module found for the frame: %s", _get_frame_repr(frame))
        return "Module name not found"
    except Exception as err:
        _SAVEFRAME_LOGGER.warning(
            "Module name couldn't be found for the frame: %s. Error: %a",
            _get_frame_repr(frame), err)
        return "Module name not found"


def _get_frame_code_line(frame):
    """
    Get the code line of the ``frame``.

    :param frame:
      The frame object.
    :return:
      The code line as returned by the `linecache` package.
    """
    filename = frame.f_code.co_filename
    lineno = frame.f_lineno
    code_line = linecache.getline(filename, lineno).strip()
    if code_line is None:
        code_line = f"No code content found at {filename!a}: {lineno}"
        _SAVEFRAME_LOGGER.info(code_line + f" for frame {_get_frame_repr(frame)}")
    return code_line


def _get_frame_metadata(frame_idx, frame_obj):
    """
    Get metadata for the frame ``frame_obj``.

    :param frame_idx:
      Index of the frame ``frame_obj`` from the bottom of the stack trace.
    :param frame_obj:
      The frame object for which to get the metadata.
    :return:
      A `FrameMetadata` object.
    """
    frame_function_object = _get_frame_function_object(frame_obj)
    try:
        if isinstance(frame_function_object, str):
            # Function object couldn't be found.
            pickled_function = frame_function_object
        else:
            pickled_function = pickle.dumps(
                frame_function_object, protocol=PICKLE_PROTOCOL)
    except Exception as err:
        _SAVEFRAME_LOGGER.info(
            "Cannot pickle the function object for the frame: %s. Error: %a",
            _get_frame_repr(frame_obj), err)
        pickled_function = "Function object not pickleable"
    # Object that stores all the frame's metadata.
    frame_metadata = FrameMetadata(
        frame_index=frame_idx,
        filename=frame_obj.f_code.co_filename,
        lineno=frame_obj.f_lineno,
        function_name=frame_obj.f_code.co_name,
        function_qualname=_get_qualname(frame_obj),
        function_object=pickled_function,
        module_name=_get_frame_module_name(frame_obj),
        code=_get_frame_code_line(frame_obj),
        frame_identifier=(
            f"{frame_obj.f_code.co_filename},{frame_obj.f_lineno},"
            f"{frame_obj.f_code.co_name}")
    )
    return frame_metadata

def _get_all_matching_frames(frame, all_frames):
    """
    Get all the frames from ``all_frames`` that match the ``frame``.

    The matching is done based on the filename / file regex, the line number
    and the function name.

    :param frame:
      Frame for which to find all the matching frames.
    :param all_frames:
      A list of all the frame objects from the exception object.
    :return:
      A list of all the frames that match the ``frame``. Each item in the list
      is a tuple of 2 elements, where the first element is the frame index
      (starting from the bottom of the stack trace) and the second element is
      the frame object.
    """
    if frame == ['']:
        # This is the case where the last frame is not passed in the range
        # ('first_frame..'). Return the first frame from the bottom as the last
        # frame.
        return [(1, all_frames[0])]
    all_matching_frames = []
    filename_regex, lineno, func_name = frame
    for idx, frame_obj in enumerate(all_frames):
        if re.search(filename_regex, frame_obj.f_code.co_filename) is None:
            continue
        if lineno and frame_obj.f_lineno != lineno:
            continue
        if (func_name and
                func_name not in (frame_obj.f_code.co_name, _get_qualname(frame_obj))):
            continue
        all_matching_frames.append((idx+1, frame_obj))
    return all_matching_frames


def _get_frames_to_save(frames, all_frames):
    """
    Get the frames we want to save from ``all_frames`` as per ``frames``.

    :param frames:
      Frames that user wants to save. This parameter stores the parsed frames
      returned by `_validate_frames` function.
    :param all_frames:
      A list of all the frame objects from the exception object.
    :return:
      A list of filtered frames to save. Each item in the list is a tuple of 2
      elements, where the first element is the frame index (starting from the
      bottom of the stack trace) and the second element is the frame object.
    """
    frames, frame_type = frames
    filtered_frames = []
    if frame_type is None:
        # No frame passed by the user, return the first frame from the bottom
        # of the stack trace.
        return [(1, all_frames[0])]
    elif frame_type == FrameFormat.NUM:
        if len(all_frames) < frames:
            _SAVEFRAME_LOGGER.info(
                "Number of frames to dump are %s, but there are only %s frames "
                "in the error stack. So dumping all the frames.",
                frames, len(all_frames))
            frames = len(all_frames)
        return [(idx+1, all_frames[idx]) for idx in range(frames)]
    elif frame_type == FrameFormat.LIST:
        for frame in frames:
            filtered_frames.extend(_get_all_matching_frames(frame, all_frames))
    elif frame_type == FrameFormat.RANGE:
        # Handle 'first_frame..last_frame' and 'first_frame..' formats.
        # Find all the matching frames for the first_frame and last_frame.
        first_matching_frames = _get_all_matching_frames(frames[0], all_frames)
        if len(first_matching_frames) == 0:
            raise ValueError(f"No frame in the traceback matched the frame: "
                             f"{':'.join(map(str, frames[0]))!a}")
        last_matching_frames = _get_all_matching_frames(frames[1], all_frames)
        if len(last_matching_frames) == 0:
            raise ValueError(f"No frame in the traceback matched the frame: "
                             f"{':'.join(map(str, frames[1]))!a}")
        # Take out the minimum and maximum indexes of the matching frames.
        first_idxs = (first_matching_frames[0][0], first_matching_frames[-1][0])
        last_idxs = (last_matching_frames[0][0], last_matching_frames[-1][0])
        # Find the maximum absolute distance between the start and end matching
        # frame indexes, and get all the frames in between that range.
        distances = [
            (abs(first_idxs[0] - last_idxs[0]), (first_idxs[0], last_idxs[0])),
            (abs(first_idxs[0] - last_idxs[1]), (first_idxs[0], last_idxs[1])),
            (abs(first_idxs[1] - last_idxs[0]), (first_idxs[1], last_idxs[0])),
            (abs(first_idxs[1] - last_idxs[1]), (first_idxs[1], last_idxs[1])),
        ]
        _, max_distance_pair = max(distances, key=lambda x: x[0])
        max_distance_pair = sorted(max_distance_pair)
        for idx in range(max_distance_pair[0], max_distance_pair[1] + 1):
            filtered_frames.append((idx, all_frames[idx-1]))

    # Only keep the unique frames and sort them using their index.
    filtered_frames = sorted(filtered_frames, key=lambda f: f[0])
    seen_frames = set()
    unique_filtered_frames = []
    # In chained exceptions, we can get the same frame at multiple indexes, so
    # get the unique frames without considering the index.
    for frame in filtered_frames:
        if frame[1] not in seen_frames:
            unique_filtered_frames.append(frame)
            seen_frames.add(frame[1])
    return unique_filtered_frames


def _get_all_frames_from_exception_obj(exception_obj):
    """
    Get all the frame objects from the exception object. It also handles chained
    exceptions.

    :param exception_obj:
      The exception raise by the user's code.
    :return:
      A list containing all the frame objects from the exception. The frames
      are stored in bottom-to-top order, with the bottom frame (the error frame)
      at index 0.
    """
    current_exception = exception_obj
    all_frames = []
    while current_exception:
        traceback = current_exception.__traceback__
        current_tb_frames = []
        while traceback:
            current_tb_frames.append(traceback.tb_frame)
            traceback = traceback.tb_next
        all_frames.extend(reversed(current_tb_frames))
        current_exception = (current_exception.__cause__ or
                             current_exception.__context__)
    return all_frames


def _save_frames_and_exception_info_to_file(
        filename, frames, variables, exclude_variables, exception_obj):
    """
    Save the frames and exception information in the file ``filename``.

    The data structure that gets saved is a dictionary. It stores each frame
    info in a separate entry with the key as the frame index (from the bottom of
    the stack trace). It also stores some useful exception information. The data
    gets dumped in the ``filename`` in pickled form. Following is the same
    structure of the info saved:
      {
          # 5th frame from the bottom
          5: {
                'frame_index': 5,
                'filename': '/path/to/file.py',
                'lineno': 3423,
                'function_name': 'func1',
                'function_qualname': 'FooClass.func1',
                'function_object': <pickled object>,
                'module_name': '<frame_module>'
                'frame_identifier': '/path/to/file.py,3423,func1',
                'code': '... python code line ...'
                'variables': {'local_variable1': <pickled value>, 'local_variable2': <pickled value>, ...}
            },
          # 17th frame from the bottom
          17: {
                'frame_index': 17,
                ...
            },
          ...
          'exception_full_string': f'{exc.__class.__name__}: {exc}'
          'exception_object': exc,
          'exception_string': str(exc),
          'exception_class_name': exc.__class__.__name__,
          'exception_class_qualname': exc.__class__.__qualname__,
          'traceback': '(multiline traceback)
      }

    :param filename:
      The file path in which to save the information.
    :param frames:
      The frames to save in the file. This parameter stores the parsed frames
      returned by the `_validate_frames` function.
    :param variables:
      The local variables to include in each frame.
    :param exclude_variables:
      The local variables to exclude from each frame.
    :param exception_obj:
      The ``Exception`` raised by the user's code. This is used to extract all
      the required info; the traceback, all the frame objects, etc.
    """
    # Mapping that stores all the information to save.
    frames_and_exception_info = {}
    # Get the list of frame objects from the exception object.
    all_frames = _get_all_frames_from_exception_obj(exception_obj)
    # Take out the frame objects we want to save as per 'frames'.
    frames_to_save = _get_frames_to_save(frames, all_frames)
    _SAVEFRAME_LOGGER.info(
        "Number of frames that'll be saved: %s", len(frames_to_save))

    for frame_idx, frame_obj in frames_to_save:
        _SAVEFRAME_LOGGER.info(
            "Getting required info for the frame: %s", _get_frame_repr(frame_obj))
        frames_and_exception_info[frame_idx] = _get_frame_metadata(
            frame_idx, frame_obj).__dict__
        frames_and_exception_info[frame_idx]['variables'] = (
            _get_frame_local_variables_data(frame_obj, variables, exclude_variables))

    _SAVEFRAME_LOGGER.info("Getting exception metadata info.")
    frames_and_exception_info.update(_get_exception_info(exception_obj).__dict__)
    _SAVEFRAME_LOGGER.info("Saving the complete data in the file: %a", filename)
    with _open_file(filename, 'wb') as f:
        pickle.dump(frames_and_exception_info, f, protocol=PICKLE_PROTOCOL)
    _SAVEFRAME_LOGGER.info("Done!!")


def _is_dir_and_ancestors_world_traversable(directory):
    """
    Is the ``directory`` and all its ancestors world traversable.

    For world traversability we check if the execute bit is set for the
    owner, group and others.

    :param directory:
      The directory to check.
    :return:
      `True` if ``directory`` and its ancestors are world traversable, else
      `False`.
    """
    required_mode = stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH
    dir_permission_mode = os.stat(directory).st_mode & 0o777
    if (dir_permission_mode & required_mode) != required_mode:
        return False
    return directory == "/" or _is_dir_and_ancestors_world_traversable(
        os.path.dirname(directory))


def _validate_filename(filename, utility):
    """
    Validate the value of ``filename``.

    :param filename:
      The file path to validate.
    :param utility:
      Indicates whether this helper is invoked by the ``pyflyby.saveframe`` function
      or the ``pyflyby/bin/saveframe`` script. See `_validate_saveframe_arguments`
      for more info.
    :return:
      The file path post validation. If ``filename`` is None, a default file
      named `DEFAULT_FILENAME` in the current working directory is returned.
    """
    if filename is None:
        filename = os.path.abspath(DEFAULT_FILENAME)
        _SAVEFRAME_LOGGER.info(
            "Filename is not passed explicitly using the %s. The frame info will "
            "be saved in the file: %a.",
            '`filename` parameter' if utility == 'function' else '--filename argument',
            filename)
    _SAVEFRAME_LOGGER.info("Validating filename: %a", filename)
    # Resolve any symlinks.
    filename = os.path.realpath(filename)
    if os.path.islink(filename):
        raise ValueError(f"Cyclic link exists in the file: {filename!a}")
    if os.path.isdir(filename):
        raise ValueError(f"{filename!a} is an already existing directory. Please "
                         f"pass a different filename.")
    if os.path.exists(filename):
        _SAVEFRAME_LOGGER.info(
            "File %a already exists. This run will overwrite the file.", filename)
    parent_dir = os.path.dirname(filename)
    # Check if the parent directory and the ancestors are world traversable.
    # Log a warning if not. Raise an error if the parent or any ancestor
    # directory doesn't exist.
    try:
        is_parent_and_ancestors_world_traversable = (
            _is_dir_and_ancestors_world_traversable(directory=parent_dir))
    except PermissionError:
        is_parent_and_ancestors_world_traversable = False
    except (FileNotFoundError, NotADirectoryError) as err:
        msg = (f"Error while saving the frames to the file: "
               f"{filename!a}. Error: {err!a}")
        raise type(err)(msg) from None
    except OSError as err:
        is_parent_and_ancestors_world_traversable = False
        _SAVEFRAME_LOGGER.warning(
            "Error while trying to determine if the parent directory: %a and "
            "the ancestors are world traversable. Error: %a", parent_dir, err)
    if not is_parent_and_ancestors_world_traversable:
        _SAVEFRAME_LOGGER.warning(
            "The parent directory %a or an ancestor is not world traversable "
            "(i.e., the execute bit of one of the ancestors is 0). The filename "
            "%a might not be accessible by others.", parent_dir, filename)
    return filename


def _validate_frames(frames, utility):
    """
    Validate the value of ``frames``.

    This utility validates / parses the ``frames`` based on the following formats:
      1. Single frame: frames=frame or frames=['frame']
      2. Multiple frames: frames='frame1,frame2,...' (for `pyflyby/bin/saveframe`
         script) or frames=['frame1', 'frame2', ...] (for `saveframe` function).
      3. Range of frames: frames='first_frame..last_frame'
      4. Range from first frame to bottom: frames='first_frame..'
      5. Number of frames from bottom: frames=num

    NOTE: Each frame is represented as 'file_regex:lineno:function_name'

    :param frames:
      Frames to validate
    :param utility:
      Indicates whether this helper is invoked by the ``pyflyby.saveframe`` function
      or the ``pyflyby/bin/saveframe`` script. See `_validate_saveframe_arguments`
      for more info.
    :return:
      A tuple of 2 items:
      1. Parsed frames:
          - None if ``frames`` is None.
          - If ``frames=num`` (case 5), this is the integer ``num``.
          - Otherwise, this is a list of tuples. Each tuple represents a frame
            and consists of three items: ``filename``, ``lineno``, and ``function_name``.
          - Example: For ``frames='/some/foo.py:32:,/some/bar.py:28:func'``, the
            first item would be ``[('/some/foo.py', 32, ''), ('/some/bar.py', 28, 'func')]``.

      2. The format of the ``frames``:
           - None if ``frames`` is None.
           - For cases 1 and 2, the format is `FrameFormat.LIST`.
           - For cases 3 and 4, the format is `FrameFormat.RANGE`.
           - For case 5, the format is `FrameFormat.NUM`.
    """
    if frames is None:
        _SAVEFRAME_LOGGER.info(
            "%s is not passed explicitly. The first frame from the bottom will be "
            "saved by default.",
            '`frames` parameter' if utility == 'function' else '--frames argument')
        return None, None
    _SAVEFRAME_LOGGER.info("Validating frames: %a", frames)
    try:
        # Handle frames as an integer.
        return int(frames), FrameFormat.NUM
    except (ValueError, TypeError):
        pass
    # Boolean to denote if the `frames` parameter is passed in the range format.
    is_range = False
    if isinstance(frames, str) and ',' in frames and utility == 'function':
        raise ValueError(
            f"Error while validating frames: {frames!a}. If you want to pass multiple "
            f"frames, pass a list/tuple of frames like {frames.split(',')} rather "
            f"than a comma separated string of frames.")
    if isinstance(frames, (list, tuple)):
        for frame in frames:
            if ',' in frame:
                raise ValueError(
                    f"Invalid frame: {frame!a} in frames: {frames} as it "
                    f"contains character ','. If you are trying to pass multiple "
                    f"frames, pass them as separate items in the list.")
        frames = ','.join(frames)
    all_frames = [frame.strip() for frame in frames.split(',')]
    # Handle the single frame and the range of frame formats.
    if len(all_frames) == 1:
        all_frames = [frame.strip() for frame in frames.split('..')]
        if len(all_frames) > 2:
            raise ValueError(
                f"Error while validating frames: {frames!a}. If you want to pass a "
                f"range of frames, the correct syntax is 'first_frame..last_frame'")
        elif len(all_frames) == 2:
            is_range = True
        else:
            is_range = False

    parsed_frames = []
    for idx, frame in enumerate(all_frames):
        frame_parts = frame.split(':')
        # Handle 'first_frame..' format (case 4.).
        if idx == 1 and len(frame_parts) == 1 and frame_parts[0] == '' and is_range:
            parsed_frames.append(frame_parts)
            break
        if len(frame_parts) != 3:
            raise ValueError(
                f"Error while validating frame: {frame!a}. The correct syntax for a "
                f"frame is 'file_regex:line_no:function_name' but frame {frame!a} "
                f"contains {len(frame_parts)-1} ':'.")
        if not frame_parts[0]:
            raise ValueError(
                f"Error while validating frame: {frame!a}. The filename / file "
                f"regex must be passed in a frame.")
        # Validate the line number passed in the frame.
        if frame_parts[1]:
            try:
                frame_parts[1] = int(frame_parts[1])
            except ValueError:
                raise ValueError(f"Error while validating frame: {frame!a}. The "
                                f"line number {frame_parts[1]!a} can't be "
                                f"converted to an integer.")
        parsed_frames.append(frame_parts)

    return parsed_frames, FrameFormat.RANGE if is_range else FrameFormat.LIST


def _is_variable_name_valid(name):
    """
    Is ``name`` a valid variable name.

    :param name:
      Variable name to validate.
    :return:
      `True` or `False`.
    """
    if not name.isidentifier():
        return False
    if keyword.iskeyword(name):
        return False
    return True


def _validate_variables(variables, utility):
    """
    Validate the value of ``variables``.

    If there are invalid variable names, filter them out and log a warning.

    :param variables:
      Variables to validate
    :param utility:
      Indicates whether this helper is invoked by the ``pyflyby.saveframe`` function
      or the ``pyflyby/bin/saveframe`` script. See `_validate_saveframe_arguments`
      for more info.
    :return:
      A tuple of filtered variables post validation.
    """
    if variables is None:
        return
    _SAVEFRAME_LOGGER.info("Validating variables: %a", variables)
    if isinstance(variables, str) and ',' in variables and utility == 'function':
        raise ValueError(
            f"Error while validating variables: {variables!a}. If you want to "
            f"pass multiple variable names, pass a list/tuple of names like "
            f"{variables.split(',')} rather than a comma separated string of names.")
    if isinstance(variables, (list, tuple)):
        all_variables = tuple(variables)
    elif isinstance(variables, str):
        all_variables = tuple(variable.strip() for variable in variables.split(','))
    else:
        raise TypeError(
            f"Variables '{variables}' must be of type list, tuple or string (for a "
            f"single variable), not '{type(variables)}'")
    invalid_variable_names = [variable for variable in all_variables
                              if not _is_variable_name_valid(variable)]
    if invalid_variable_names:
        _SAVEFRAME_LOGGER.warning(
            "Invalid variable names: %s. Skipping these variables and continuing.",
            invalid_variable_names)
        # Filter out invalid variables.
        all_variables = tuple(variable for variable in all_variables
                              if variable not in invalid_variable_names)
    return all_variables


def _validate_saveframe_arguments(
        filename, frames, variables, exclude_variables, utility='function'):
    """
    Validate and sanitize the parameters supported by the `saveframe` function.

    :param filename:
      File path in which to save the frame's info.
    :param frames:
      Specific error frames to save.
    :param variables:
      Local variables to include in each frame info.
    :param exclude_variables:
      Local variables to exclude from each frame info.
    :param utility:
      Indicates whether this helper is invoked by the ``pyflyby.saveframe`` function
      or the ``pyflyby/bin/saveframe`` script. Allowed values are 'function' and
      'script'. The saveframe function and script accept different types of
      values for their arguments. This parameter helps distinguish between them
      to ensure proper argument and parameter validation.
    :return:
      A tuple of ``filename``, ``frames``, ``variables`` and ``exclude_variables``
      post validation.
    """
    allowed_utility_values = ['function', 'script']
    if utility not in allowed_utility_values:
        raise ValueError(
            f"Invalid value for parameter 'utility': {utility!a}. Allowed values "
            f"are: {allowed_utility_values}")
    filename = _validate_filename(filename, utility)
    frames =_validate_frames(frames, utility)
    if variables and exclude_variables:
        raise ValueError(
            f"Cannot pass both {'`variables`' if utility == 'function' else '--variables'} "
            f"and {'`exclude_variables`' if utility == 'function' else '--exclude_variables'} "
            f"{'parameters' if utility == 'function' else 'arguments'}.")
    variables = _validate_variables(variables, utility)
    exclude_variables = _validate_variables(exclude_variables, utility)
    if not (variables or exclude_variables):
        _SAVEFRAME_LOGGER.info(
            "Neither %s nor %s %s is passed. All the local variables from the "
            "frames will be saved.",
            '`variables`' if utility == 'function' else '--variables',
            '`exclude_variables`' if utility == 'function' else '--exclude_variables',
            'parameter' if utility == 'function' else 'argument')

    return filename, frames, variables, exclude_variables


def saveframe(filename=None, frames=None, variables=None, exclude_variables=None):
    """
    Utility to save information for debugging / reproducing an issue.

    Usage:
    --------------------------------------------------------------------------
    If you have a piece of code that is currently failing due to an issue
    originating from upstream code, and you cannot share your private
    code as a reproducer, use this function to save relevant information to a file.
    While in an interactive session such as IPython, Jupyter Notebook, or a
    debugger (pdb/ipdb), you can call this function after your code raised
    an error to capture and save error frames specific to the upstream codebase
    Share the generated file with the upstream team, enabling them to reproduce
    and diagnose the issue independently.

    Information saved in the file:
    --------------------------------------------------------------------------
    This utility captures and saves error stack frames to a file. It includes the
    values of local variables from each stack frame, as well as metadata about each
    frame and the exception raised by the user's code. Following is the sample
    structure of the info saved in the file:

    ::

      {
          # 5th frame from the bottom
          5: {
                'frame_index': 5,
                'filename': '/path/to/file.py',
                'lineno': 3423,
                'function_name': 'func1',
                'function_qualname': 'FooClass.func1',
                'function_object': <pickled object>,
                'module_name': '<frame_module>'
                'frame_identifier': '/path/to/file.py,3423,func1',
                'code': '... python code line ...'
                'variables': {'local_variable1': <pickled value>, 'local_variable2': <pickled value>, ...}
            },
          # 17th frame from the bottom
          17: {
                'frame_index': 17,
                ...
            },
          ...
          'exception_full_string': f'{exc.__class.__name__}: {exc}'
          'exception_object': exc,
          'exception_string': str(exc),
          'exception_class_name': exc.__class__.__name__,
          'exception_class_qualname': exc.__class__.__qualname__,
          'traceback': '(multiline traceback)
      }

    .. note::
      - The above data gets saved in the file in pickled form.
      - In the above data, the key of each frame's entry is the index of that frame
        from the bottom of the error stack trace. So the first frame from the bottom
        (the error frame) has index 1, and so on.
      - 'variables' key in each frame's entry stores the local variables of that frame.
      - The 'exception_object' key stores the actual exception object but without
        the __traceback__ info (for security reasons).

    **Example usage**:

    ::

      # In an interactive session (ipython, jupyter notebook, etc.)

      >> <Your code raised an error>
      >> saveframe(filename=/path/to/file) # Saves the first frame from the bottom
      >> saveframe(filename=/path/to/file, frames=frames_to_save,
      .. variables=local_variables_to_save, exclude_variables=local_variables_to_exclude)

      # In an interactive debugger (pdb / ipdb)

      >> <Your code raised an error>
      >> ipdb.pm() # start a debugger
      >> ipdb> from pyflyby import saveframe
      >> ipdb> saveframe(filename=/path/to/file) # Saves the frame which you are currently at
      >> ipdb> saveframe(filename=/path/to/file, frames=frames_to_save,
      .. variables=local_variables_to_include, exclude_variables=local_variables_to_exclude)

      # Let's say your code is raising an error with the following traceback:

      File "dir/__init__.py", line 6, in init_func1
          func1()
      File "dir/mod1.py", line 14, in func1
          func2()
      File "dir/mod1.py", line 9, in func2
          obj.func2()
      File "dir/pkg1/mod2.py", line 10, in func2
          func3()
      File "dir/pkg1/pkg2/mod3.py", line 6, in func3
          raise ValueError("Error is raised")
      ValueError: Error is raised

      # To save the last frame (the error frame) in file '/path/to/file', use:
      >> saveframe(filename='/path/to/file')

      # To save a specific frame like `File "dir/mod1.py", line 9, in func2`, use:
      >> saveframe(filename='/path/to/file', frames='mod1.py:9:func2')

      # To save the last 3 frames from the bottom, use:
      >> saveframe(frames=3)

      # To save all the frames from 'mod1.py' and 'mod2.py' files, use:
      >> saveframe(filename='/path/to/file', frames=['mod1.py::', 'mod2.py::'])

      # To save a range of frames from 'mod1.py' to 'mod3.py', use:
      >> saveframe(frames='mod1.py::..mod3.py::')

      # To save a range of frames from '__init__.py' till the last frame, use:
      >> saveframe(frames='__init__.py::..')

      # To only save local variables 'var1' and 'var2' from the frames, use:
      >> saveframe(frames=<frames_to_save>, variables=['var1', 'var2'])

      # To exclude local variables 'var1' and 'var2' from the frames, use:
      >> saveframe(frames=<frames_to_save>, exclude_variables=['var1', 'var2'])

    For non-interactive use cases (e.g., a failing script or command), checkout
    `pyflyby/bin/saveframe` script.

    :param filename:
      File path in which to save the frame information. If this file already
      exists, it will be overwritten; otherwise, a new file will be created
      with permission mode '0o644'. If this parameter is not passed, the info
      gets saved in the 'saveframe.pkl' file in the current working directory.

    :param frames:
      Error stack frames to save. A single frame follows the format
      'filename:line_no:function_name', where:
        - filename: The file path or a regex pattern matching the file path
          (displayed in the stack trace) of that error frame.
        - line_no (Optional): The code line number (displayed in the stack trace)
          of that error frame.
        - function_name (Optional): The function name (displayed in the stack trace)
          of that error frame.

      Partial frames are also supported where line_no and/or function_name can
      be omitted:
        - filename:: -> Includes all the frames that matches the filename
        - filename:line_no: -> Include all the frames that matches specific line
          in any function in the filename
        - filename::function_name -> Include all the frames that matches any line
          in the specific function in the filename

      Following formats are supported to pass the frames:

        1. Single frame:
           frames='frame'
           Example: frames='/path/to/file.py:24:some_func'
           Includes only the specified frame.

        2. Multiple frames:
           frames=['frame1', 'frame2', ...]
           Example: frames=['/dir/foo.py:45:', '.*/dir2/bar.py:89:caller']
           Includes all specified frames.

        3. Range of frames:
           frames='first_frame..last_frame'
           Example: frames='/dir/foo.py:45:get_foo../dir3/blah.py:23:myfunc'
           Includes all the frames from first_frame to last_frame (both inclusive).

        4. Range from first_frame to bottom:
           frames='first_frame..'
           Example: frames='/dir/foo.py:45:get_foo..'
           Includes all the frames from first_frame to the bottom of the stack trace.

        5. Number of Frames from Bottom:
           frames=num
           Example: frames=5
           Includes the first 'num' frames from the bottom of the stack trace.

      Default behavior if this parameter is not passed:
        - When user is in a debugger (ipdb/pdb): Save the frame the user is
          currently at.
        - When user is not in a debugger: Save the first frame from the bottom
          (the error frame).

    :param variables:
      Local variables to include in each frame. It accepts a list/tuple of
      variable names or a string if there is only 1 variable.

      If this parameter is not passed, save all the local variables of the
      included frames.

    :param exclude_variables:
      Local variables to exclude from each frame. It accepts a list/tuple of
      variable names or a string if there is only 1 variable.

      If this parameter is not passed, save all the local variables of the
      included frames as per the ``variables`` parameter value.

    :return:
      The file path in which the frame info is saved.
    """
    if not ((sys.version_info < (3, 12) and hasattr(sys, 'last_value')) or
            (sys.version_info >= (3, 12) and hasattr(sys, 'last_exc'))):
        raise RuntimeError(
            "No exception is raised currently for which to save the frames. "
            "Make sure that an uncaught exception is raised before calling the "
            "`saveframe` function.")
    # Get the latest exception raised.
    exception_obj = sys.last_value if sys.version_info < (3, 12) else sys.last_exc

    if frames is None:
        # Get the instance of the interactive session the user is currently in.
        interactive_session_obj = sys._getframe().f_back.f_back.f_locals.get('self')
        # If the user is currently in a debugger (ipdb/pdb), save the frame the
        # user is currently at in the debugger.
        if interactive_session_obj and hasattr(interactive_session_obj, 'curframe'):
            current_frame = interactive_session_obj.curframe
            frames = (f"{current_frame.f_code.co_filename}:{current_frame.f_lineno}:"
                      f"{_get_qualname(current_frame)}")

    _SAVEFRAME_LOGGER.info("Validating arguments passed.")
    filename, frames, variables, exclude_variables = _validate_saveframe_arguments(
        filename, frames, variables, exclude_variables)
    _SAVEFRAME_LOGGER.info(
        "Saving frames and metadata for the exception: %a", exception_obj)
    _save_frames_and_exception_info_to_file(
        filename=filename, frames=frames, variables=variables,
        exclude_variables=exclude_variables, exception_obj=exception_obj)
    return filename



# pyflyby/test_imports2s.py
import tempfile

# License for THIS FILE ONLY: CC0 Public Domain Dedication
# http://creativecommons.org/publicdomain/zero/1.0/



import pytest
import sys
from   textwrap                 import dedent
import types

from   pyflyby._importdb        import ImportDB
from   pyflyby._imports2s       import (canonicalize_imports,
                                        fix_unused_and_missing_imports,
                                        reformat_import_statements,
                                        remove_broken_imports,
                                        replace_star_imports,
                                        transform_imports)
from   pyflyby._parse           import PythonBlock


def test_reformat_import_statements_1():
    input = PythonBlock(dedent('''
        from foo import bar2 as bar2x, bar1
        import foo.bar3 as bar3x
        import foo.bar4

        import foo.bar0 as bar0
    ''').lstrip(), filename="/foo/test_reformat_import_statements_1.py")
    output = reformat_import_statements(input)
    expected = PythonBlock(dedent('''
        import foo.bar4
        from foo import bar1, bar2 as bar2x, bar3 as bar3x

        from foo import bar0
    ''').lstrip(), filename="/foo/test_reformat_import_statements_1.py")
    assert output == expected


def test_reformat_import_statements_star_1():
    input = PythonBlock(dedent('''
        from mod1 import f1b
        from mod1 import f1a
        from mod2 import *
        from mod2 import f2b as F2B
        from mod2 import f2a as F2A
        from mod3 import f3b
        from mod3 import f3a
    ''').lstrip(), filename="/foo/test_reformat_import_statements_star_1.py")
    output = reformat_import_statements(input)
    expected = PythonBlock(dedent('''
        from mod1 import f1a, f1b
        from mod2 import *
        from mod2 import f2a as F2A, f2b as F2B
        from mod3 import f3a, f3b
    ''').lstrip(), filename="/foo/test_reformat_import_statements_star_1.py")
    assert output == expected


def test_reformat_import_statements_multi_star_1():
    input = PythonBlock(dedent('''
        from mod1 import *
        from mod2 import *
    ''').lstrip())
    output = reformat_import_statements(input)
    expected = PythonBlock(dedent('''
        from mod1 import *
        from mod2 import *
    ''').lstrip())
    assert output == expected


def test_reformat_import_statements_shadowed_1():
    input = PythonBlock(dedent('''
        import a, a2 as a, b, b2 as b, b, c2 as c, d2 as d
        from d import d
        import c
    ''').lstrip())
    output = reformat_import_statements(input)
    expected = PythonBlock(dedent('''
        import a2 as a
        import b
        import c
        from d import d
    ''').lstrip())
    assert output == expected


def test_reformat_continuation_comments_1():
    input = dedent('''
        import   b , a
        # x\\
        # y
    ''').lstrip()
    output = str(reformat_import_statements(input))
    expected = dedent('''
        import a
        import b
        # x\\
        # y
    ''').lstrip()
    assert expected == output


def test_fix_unused_and_missing_imports_1():
    input = PythonBlock(dedent('''
        from foo import m1, m2, m3, m4
        m2, m4, np.foo
    ''').lstrip(), filename="/foo/test_fix_unused_and_missing_imports_1.py")
    db = ImportDB("""
        import numpy as np
        __mandatory_imports__ = ["from __future__ import division"]
    """)
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from __future__ import division

        import numpy as np
        from foo import m2, m4
        m2, m4, np.foo
    ''').lstrip(), filename="/foo/test_fix_unused_and_missing_imports_1.py")
    assert output == expected


def test_fix_unused_and_missing_imports_unknown_1():
    input = PythonBlock(dedent('''
        os, sys
    ''').lstrip())
    db = ImportDB("import os")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        import os

        os, sys
    ''').lstrip())
    assert output == expected


def test_fix_unused_and_missing_imports_known_different_1():
    input = PythonBlock(dedent('''
        from ma import a1, a2
        from mb import b1, b2
        a1, a2, a3, b1, b2, b3
    ''').lstrip())
    db = ImportDB("from MA import a1, a2, a3\nfrom MB import b1, b2, b3\n")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from MA import a3
        from MB import b3
        from ma import a1, a2
        from mb import b1, b2
        a1, a2, a3, b1, b2, b3
    ''').lstrip())
    assert output == expected


def test_fix_unused_and_missing_imports_insertion_heuristic_1():
    input = PythonBlock(dedent('''
        from m1.a import f1
        from m2.a import f2
        from m3.a import f3
        f1, f2, f3, f4
    ''').lstrip())
    db = ImportDB("from m2.b import f4")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from m1.a import f1
        from m2.a import f2
        from m2.b import f4
        from m3.a import f3
        f1, f2, f3, f4
    ''').lstrip())
    assert output == expected


def test_fix_unused_and_missing_imports_insertion_after_comments_1():
    input = PythonBlock(dedent('''
        # hello

        # there
        f1, f2
    ''').lstrip())
    db = ImportDB("from m1 import f1")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        # hello

        # there
        from m1 import f1

        f1, f2
    ''').lstrip())
    assert output == expected


def test_fix_unused_and_missing_imports_insertion_after_docstring_1():
    input = PythonBlock(dedent('''
        """
        aaa
        """
        f1, f2
    ''').lstrip())
    db = ImportDB("from m1 import f1")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        """
        aaa
        """
        from m1 import f1

        f1, f2
    ''').lstrip())
    assert output == expected


def test_fix_unused_and_missing_imports_insertion_after_docstring_cont_1():
    input = PythonBlock(dedent('''
        """
        aaa
        """ '\
        bbb'
        f1, f2
    ''').lstrip())
    db = ImportDB("from m1 import f1")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        """
        aaa
        """ '\
        bbb'
        from m1 import f1

        f1, f2
    ''').lstrip())
    assert output == expected


def test_fix_unused_and_missing_imports_insertion_after_docstring_whitespace_1():
    input = PythonBlock(dedent('''
        """
        aaa
        """


        f1, f2
    ''').lstrip())
    db = ImportDB("from m1 import f1")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        """
        aaa
        """


        from m1 import f1

        f1, f2
    ''').lstrip())
    assert output == expected


def test_fix_unused_and_missing_imports_insertion_comments_1():
    input = PythonBlock(dedent('''
        #c0
        from m1.y import m1yf
        #c1
        from m2.y import m2yf
        #c2
        from m3.y import m3yf
        #c3
        m1xf, m2xf, m3xf
        m1yf, m2yf, m3yf
        m1zf, m2zf, m3zf
    ''').lstrip())
    db = ImportDB("""
        from m1.x import m1xf
        from m1.y import m1yf
        from m1.z import m1zf
        from m2.x import m2xf
        from m2.y import m2yf
        from m2.z import m2zf
        from m3.x import m3xf
        from m3.y import m3yf
        from m3.z import m3zf
    """)
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        #c0
        from m1.x import m1xf
        from m1.y import m1yf
        from m1.z import m1zf
        #c1
        from m2.x import m2xf
        from m2.y import m2yf
        from m2.z import m2zf
        #c2
        from m3.x import m3xf
        from m3.y import m3yf
        from m3.z import m3zf
        #c3
        m1xf, m2xf, m3xf
        m1yf, m2yf, m3yf
        m1zf, m2zf, m3zf
    ''').lstrip())
    assert output == expected


def test_fix_unused_and_missing_imports_compound_statements_1():
    input = PythonBlock(dedent('''
        import a, b; import c, d; g
        a, c; f
    ''').lstrip())
    db = ImportDB("""
        from m1 import f
        from m1 import g
    """)
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        import a
        import c
        from m1 import f, g
        g
        a, c; f
    ''').lstrip())
    assert output == expected


def test_fix_unused_and_missing_imports_multiple_import_used_1():
    input = PythonBlock(dedent('''
        from m0 import f, x
        import z, x
        import y, z
        from m1 import h, f
        import x
        import x
        from m1 import f, f, g
        x, f
    ''').lstrip())
    db = ImportDB("")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        import x
        from m1 import f
        x, f
    ''').lstrip())
    assert output == expected


def test_fix_unused_and_missing_imports_shadowed_1():
    input = PythonBlock(dedent('''
        from m1 import f1, f2, f3
        def f2(): f1
    ''').lstrip())
    db = ImportDB("")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from m1 import f1
        def f2(): f1
    ''').lstrip())
    assert output == expected


def test_fix_unused_and_missing_imports_reassigned_first_reference_1():
    input = PythonBlock(dedent('''
        from bar import foo1, foo2
        foo2 = foo2 + 0
    ''').lstrip())
    db = ImportDB("")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from bar import foo2
        foo2 = foo2 + 0
    ''').lstrip())
    assert output == expected


def test_fix_unused_and_missing_imports_midfile_1():
    input = PythonBlock(dedent('''
        import m1, m2
        m2, m3
        import m4, m5
        m4, m6
    ''').lstrip())
    db = ImportDB("import m1, m2, m3, m4, m5, m6, m7")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        import m2
        import m3
        m2, m3
        import m4
        import m6
        m4, m6
    ''').lstrip())
    assert output == expected


def test_fix_unused_and_missing_imports_doctests_1():
    input = PythonBlock(dedent('''
        from m1 import f1, f2, f3
        def foo():
            """
              >>> f1 + f2 + f9
            """
            return None
    ''').lstrip())
    db = ImportDB("")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from m1 import f1, f2
        def foo():
            """
              >>> f1 + f2 + f9
            """
            return None
    ''').lstrip())
    assert output == expected


@pytest.mark.xfail
def test_fix_unused_and_missing_imports_doctests_2():
    input = PythonBlock(dedent('''
        from m1 import f1, f2, f3
        def foo():
            """
              >>> f1 = 0
              >>> f1 + f2 + f9
            """
            return None
    ''').lstrip())
    db = ImportDB("")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from m1 import f2
        def foo():
            """
              >>> f1 = 0
              >>> f1 + f2 + f9
            """
            return None
    ''').lstrip())
    assert output == expected


def test_fix_unused_and_missing_imports_xref_1():
    input = PythonBlock(dedent('''
        from m1 import f1, f2, f3
        def foo():
            """
            Hello L{f1} C{f3}
            """
            return None
    ''').lstrip())
    db = ImportDB("")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from m1 import f1, f3
        def foo():
            """
            Hello L{f1} C{f3}
            """
            return None
    ''').lstrip())
    assert output == expected


def test_fix_unused_and_missing_imports_decorator_2():
    input = PythonBlock(dedent('''
        from m1 import d1, d2, d3
        @d1
        def f1(): pass
        @d9.d2
        def f9(): pass
    ''').lstrip())
    db = ImportDB("from m2 import d8, d9")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from m1 import d1
        from m2 import d9
        @d1
        def f1(): pass
        @d9.d2
        def f9(): pass
    ''').lstrip())
    assert output == expected


def test_fix_unused_imports_funcall_1():
    input = PythonBlock(dedent('''
        from m1 import X1, X2
        def F1(X1): X1, X2
    ''').lstrip())
    db = ImportDB("")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from m1 import X2
        def F1(X1): X1, X2
    ''').lstrip())
    assert output == expected


def test_fix_unused_and_missing_imports_funcall_1():
    input = PythonBlock(dedent('''
        from m1 import X1, X3, X9
        def F1(X1, X2=x2, X3=x3, *X4, **X5): X1, X5, x6, X9
        def F2(X7=x7): X7
        def F3(a1): a1
        def F4(): a1
    ''').lstrip())
    db = ImportDB("from m2 import x1, x2, x3, x4, x5, x6, x7, a1, a2")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from m1 import X9
        from m2 import a1, x2, x3, x6, x7
        def F1(X1, X2=x2, X3=x3, *X4, **X5): X1, X5, x6, X9
        def F2(X7=x7): X7
        def F3(a1): a1
        def F4(): a1
    ''').lstrip())
    assert output == expected


def test_fix_missing_imports_funcall_1():
    input = PythonBlock(dedent('''
        def F1(x1): x1, x2, y1
    ''').lstrip())
    db = ImportDB("from m2 import x1, x2, x3, x4")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from m2 import x2

    ''').lstrip() + str(input))
    assert output == expected


def test_fix_missing_imports_funcall_and_really_missing_1():
    input = PythonBlock(dedent('''
        def F1(x1, x2, x3): x1, x2, x3, x4, y1
        x1
    ''').lstrip())
    db = ImportDB("from m2 import x1, x2, x3, x4, x5")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from m2 import x1, x4

    ''').lstrip() + str(input))
    assert output == expected


def test_fix_unused_imports_local_1():
    input = PythonBlock(dedent('''
        from m1 import X1, X2
        def F1():
            X1 = 5
            X1, X2
    ''').lstrip())
    db = ImportDB("")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from m1 import X2
        def F1():
            X1 = 5
            X1, X2
    ''').lstrip())
    assert output == expected


def test_fix_missing_imports_local_1():
    input = PythonBlock(dedent('''
        def F1():
            x1 = 5
            x1, x2
    ''').lstrip())
    db = ImportDB("from m2 import x1, x2, x3")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from m2 import x2

        def F1():
            x1 = 5
            x1, x2
    ''').lstrip())
    assert output == expected


def test_fix_missing_imports_nonlocal_post_store_1():
    input = PythonBlock(dedent('''
        def F1():
            x1 = None
            def F2():
                x1, x2, x3, y1
            x2 = None
    ''').lstrip())
    db = ImportDB("from m2 import x1, x2, x3, x4, x5")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from m2 import x3

    ''').lstrip() + str(input))
    assert output == expected


def test_fix_missing_imports_nonlocal_post_del_1():
    """
    Calling F2 after the deletion of x2 in the enclosing scope make no sens, it
    would trigger a:

    NameError: free variable 'x2' referenced before assignment in enclosing scope

    Regardless as to whether x2 is defines at module scope or not.

    Therefore we do not expect x2 to be imported
    """
    input = PythonBlock(dedent('''
        def F1():
            x1 = x2 = None
            def F2():
                x1, x2, x3, y1
            del x2
    ''').lstrip())
    db = ImportDB("from m2 import x1, x2, x3, x4")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from m2 import x3

    ''').lstrip() + str(input))
    assert output == expected


def test_fix_unused_and_missing_imports_IfExp_1():
    input = PythonBlock(dedent('''
        x if y else z
    ''').lstrip())
    db = ImportDB("from m1 import w, x, y, z")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from m1 import x, y, z

        x if y else z
    ''').lstrip())
    assert output == expected


def test_fix_unused_and_missing_imports_ClassDef_1():
    input = PythonBlock(dedent('''
        @x1
        class x2(x3(x4), x5):
            @x6
            def x7(x8=x9): x10, x8
            def x11(): x11
    ''').lstrip())
    db = ImportDB("from m1 import x1,x2,x3,x4,x5,x6,x7,x8,x9,x10,x11,x12")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from m1 import x1, x10, x11, x3, x4, x5, x6, x9

        @x1
        class x2(x3(x4), x5):
            @x6
            def x7(x8=x9): x10, x8
            def x11(): x11
    ''').lstrip())
    assert output == expected


def test_fix_missing_imports_in_non_method():
    """
    unlike test_fix_unused_and_missing_imports_ClassDef_1
    Free standing functions can refer to themselves.

    Currently this will not work for closure defined in
    methods, as we'll see a class scope.

    See https://github.com/deshaw/pyflyby/issues/179
    """
    input = PythonBlock(
        dedent(
            """
            def selfref():
                selfref.execute = True
    """
        ).lstrip()
    )
    db = ImportDB("")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(
        dedent(
            """
            def selfref():
                selfref.execute = True
    """
        ).lstrip()
    )
    assert output == expected


@pytest.mark.parametrize('data',("""
    from __future__ import annotations

    class _TestCls():
        class WrappedFunction():
            pass

    class Popen():
        def test(self) -> Popen:
            return None""",
    """
    from __future__ import annotations

    class _TestCls():
        class WrappedFunction():
            pass

    class Popen():
        def test(self:Popen):
            return None
    """))
def test_fix_missing_imports_other_class(data):
    """
    This insert import for type annotations, only if a class is present before.

    See https://github.com/deshaw/pyflyby/issues/318
    """
    data = dedent(data)
    b = PythonBlock(data)
    db = ImportDB("from subprocess import Popen")
    output = fix_unused_and_missing_imports(b, db=db)
    expected = PythonBlock(data)
    assert output == expected


def test_fix_unused_and_missing_continutation_1():
    input = PythonBlock(dedent(r'''
        a#\
        b + '\
        c#' + d
    ''').lstrip())
    db = ImportDB("from m1 import a, b, c, d")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent(r'''
        from m1 import a, b, d

        a#\
        b + '\
        c#' + d
    ''').lstrip())
    assert output == expected


def test_fix_unused_import_future_is_not_unused_1():
    input = PythonBlock(dedent(r'''
        from __future__ import division
    ''').lstrip())
    db = ImportDB("")
    output = fix_unused_and_missing_imports(input, db=db)
    assert output == input


def test_fix_unused_and_missing_print_function_1():
    input = PythonBlock(dedent(r'''
        from __future__ import print_function
        from m1 import print, m1a
        from m2.print import m2a, m2b
        m1a, m2a, m3a
    ''').lstrip())
    db = ImportDB(
        "from __future__ import print_function\n"
        "from print.m3 import m3a, m3b")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent(r'''
        from __future__ import print_function
        from m1       import m1a
        from m2.print import m2a
        from print.m3 import m3a
        m1a, m2a, m3a
    ''').lstrip())
    assert output == expected


def test_last_line_no_trailing_newline_1():
    input = PythonBlock("#x\ny")
    db = ImportDB("from Y import y")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        #x
        from Y import y

        y''').lstrip())
    assert output == expected


def test_last_line_comment_no_trailing_newline_1():
    input = PythonBlock("y\n#x")
    db = ImportDB("from Y import y")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from Y import y

        y
        #x''').lstrip())
    assert output == expected


def test_last_line_multistring_no_trailing_newline_1():
    input = PythonBlock(dedent('''
        """
        #x""", y # comment  ''').lstrip())
    db = ImportDB("from Y import y")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from Y import y

        """
        #x""", y # comment  ''').lstrip())
    assert output == expected


def test_last_line_escaped_string_no_trailing_newline_1():
    input = PythonBlock(dedent('''
        "\
        #x", y # comment  ''').lstrip())
    db = ImportDB("from Y import y")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from Y import y

        "\
        #x", y # comment  ''').lstrip())
    assert output == expected


def test_remove_broken_imports_1():
    input = PythonBlock(dedent('''
        import sys, os, omgdoesntexist_95421787, keyword
        from email.mime.audio import MIMEAudio, omgdoesntexist_8824165
        code()
    ''').lstrip(), filename="/foo/test_remove_broken_imports_1.py")
    output = remove_broken_imports(input)
    expected = PythonBlock(dedent('''
        import keyword
        import os
        import sys
        from email.mime.audio import MIMEAudio
        code()
    ''').lstrip(), filename="/foo/test_remove_broken_imports_1.py")
    assert output == expected


def test_replace_star_no_imports_found(capsys):
    m = types.ModuleType("fake_test_module_345490")
    sys.modules["fake_test_module_345490"] = m
    input = PythonBlock(dedent('''
        from fake_test_module_345490 import *
    ''').lstrip(), filename="/foo/test_replace_star_imports_2.py")
    _ = replace_star_imports(input)
    captured = capsys.readouterr()
    assert 'Traceback' not in captured.err


def test_replace_star_imports_os_issue_281(capsys):
    input = PythonBlock(dedent('''
        from os import *

        getcwd()
    ''').lstrip(), filename="/foo/test_replace_os_imports.py")
    _ = replace_star_imports(input)
    captured = capsys.readouterr()
    assert 'with' in captured.out and 'imports' in captured.out


@pytest.mark.skipif(
        sys.version_info >= (3,12),
        reason='Pyton 3.12 uses find_spec which does not work with injected modules')
def test_replace_star_imports_1():
    input = PythonBlock(dedent('''
        from mod1                    import f1
        from fake_test_module_345489 import *
        from mod2                    import f5
    ''').lstrip(), filename="/foo/test_replace_star_imports_1.py")
    m = types.ModuleType("fake_test_module_345489")
    m.__all__ = ['f1', 'f2', 'f3', 'f4', 'f5']
    sys.modules["fake_test_module_345489"] = m

    with tempfile.NamedTemporaryFile(suffix=".py", mode='w+') as f:
        f.write(f"__all__ = {m.__all__}")
        m.__file__ = f.name
        f.flush()
        output = replace_star_imports(input)

    expected = PythonBlock(dedent('''
        from fake_test_module_345489 import f1, f2, f3, f4
        from mod2                    import f5
    ''').lstrip(), filename="/foo/test_replace_star_imports_1.py")
    assert output == expected


def test_replace_star_imports_relative_1():
    # Not implemented (semi-intentionally), but at least don't crash.
    input = PythonBlock(dedent('''
        from .x import *
    ''').lstrip(), filename="/foo/test_replace_star_imports_relative_1.py")
    output = replace_star_imports(input)
    expected = PythonBlock(dedent('''
        from .x import *
    ''').lstrip(), filename="/foo/test_replace_star_imports_relative_1.py")
    assert output == expected


def test_replace_star_imports_unknown_module_1():
    input = PythonBlock(dedent('''
        from omgnonexistentmodule75085477 import *
    ''').lstrip())
    output = replace_star_imports(input)
    expected = PythonBlock(dedent('''
        from omgnonexistentmodule75085477 import *
    ''').lstrip())
    assert output == expected


def test_transform_imports_1():
    input = PythonBlock(dedent('''
        from m import x
        from m import x as X
        import m.x
        print(m.x, m.xx)
    ''').lstrip(), filename="/foo/test_transform_imports_1.py")
    output = transform_imports(input, {"m.x": "m.y.z"})
    expected = PythonBlock(dedent('''
        import m.y.z
        from m.y import z as X, z as x
        print(m.y.z, m.xx)
    ''').lstrip(), filename="/foo/test_transform_imports_1.py")
    assert output == expected


def test_canonicalize_imports_1():
    input = PythonBlock(
        dedent(
            """
        from very_very_very_long_named_package.very_long_named_module \\
                                import(small_class, very_large_named_class as small_name)
        from m import x
        from m import x as X
        import m.x
        print(m.x, m.xx)
    """
        ).lstrip(),
        filename="/foo/test_transform_imports_1.py",
    )
    db = ImportDB(
        """
        __canonical_imports__ = {"m.x": "m.y.z"}
    """
    )
    output = canonicalize_imports(input, db=db)
    expected = PythonBlock(
        dedent("""
        import m.y.z
        from m.y                                                      import (z as X,
                                                                              z as x)
        from very_very_very_long_named_package.very_long_named_module import (small_class,
                                                                              very_large_named_class as small_name)
        print(m.y.z, m.xx)
        """).lstrip(),
        filename="/foo/test_transform_imports_1.py",
    )
    assert output == expected


def test_canonicalize_imports_f_string_1():
    input = PythonBlock(dedent('''
        a = 1
        print(f"{a:2d}")
    ''').lstrip(), filename="/foo/test_canonicalize_imports_f_string_1.py")
    db = ImportDB("""
    """)
    output = canonicalize_imports(input, db=db)
    expected = PythonBlock(dedent('''
        a = 1
        print(f"{a:2d}")
    ''').lstrip(), filename="/foo/test_canonicalize_imports_f_string_1.py")
    assert output == expected

def test_empty_file_1():
    input = PythonBlock('', filename="/foo/test_empty_file_1.py")
    db = ImportDB("")
    output = canonicalize_imports(input, db=db)
    expected = PythonBlock('', filename="/foo/test_empty_file_1.py")
    assert output == expected


def test_empty_file_mandatory_1():
    input = PythonBlock('', filename="/foo/test_empty_file_mandatory_1.py")
    db = ImportDB("__mandatory_imports__ = ['from aa import cc,bb']")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock('from aa import bb, cc\n\n',
                           filename="/foo/test_empty_file_mandatory_1.py")
    assert output == expected


def test_future_flags_1():
    input = PythonBlock(dedent('''
        from __future__ import print_function

        print("", file=sys.stdout)
    ''').lstrip())
    db = ImportDB("import os, sys")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from __future__ import print_function
        import sys

        print("", file=sys.stdout)
    ''').lstrip())
    assert output == expected


def test_with_1():
    input = PythonBlock(dedent('''
        with   closing(open("/etc/passwd")) as f:
            pass
    ''').lstrip())
    db = ImportDB("from contextlib import closing")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        from contextlib import closing

        with   closing(open("/etc/passwd")) as f:
            pass
    ''').lstrip())
    assert expected == output


def test_fix_unused_imports_repeated_1():
    input = PythonBlock(dedent('''
        import foo1, foo2, foo1, foo1
        import foo1, foo2, foo3
        foo1, foo2
    '''))
    db = ImportDB("")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        import foo1
        import foo2
        foo1, foo2
    '''))
    assert expected == output


def test_fix_unused_imports_dunder_file_1(capsys):
    input = PythonBlock(dedent('''
        __file__, __asdf__
    '''))
    db = ImportDB("")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        __file__, __asdf__
    '''))
    assert expected == output
    out, err = capsys.readouterr()
    assert "undefined name '__asdf__'"     in out
    assert "undefined name '__file__'" not in out
    assert not err


def test_x_shadow_comprehension(capsys):
    input = PythonBlock(
        dedent(
            """
    def f():
        x = 1
        y = [x for _ in range(5)]
        del x
    """
        )
    )
    db = ImportDB("")
    output = fix_unused_and_missing_imports(input, db=db)
    out, err = capsys.readouterr()
    assert output == input
    assert "undefined name 'x'" not in out
    assert not err

def test_fix_unused_imports_submodule_1():
    input = PythonBlock(dedent('''
        import m2.y
        m2.x + m2.y + m3.x
    '''))
    db = ImportDB("import m2, m3, m4")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        import m2.y
        import m3
        m2.x + m2.y + m3.x
    '''))
    assert expected == output


def test_fix_unused_imports_submodule_importas_1():
    input = PythonBlock(dedent('''
        import m2.y as m2y
        m2.x + m2y + m3.x
    '''))
    db = ImportDB("import m2, m3, m4")
    output = fix_unused_and_missing_imports(input, db=db)
    expected = PythonBlock(dedent('''
        import m2
        import m3
        from m2 import y as m2y
        m2.x + m2y + m3.x
    '''))
    assert expected == output



# The following __mandatory_imports__ line makes tidy-imports add the
# mentioned imports to all files.
# You can also include imports other than __future__ imports.
#
# You can write this in any *.py file in your $PYFLYBY_PATH,
# e.g. ~/.pyflyby/mandatory.py.

# __mandatory_imports__ = [
#     'from __future__ import absolute_import, division',
# ]
